#include "flang.h"
#include "eval.h"

object_t reg_task;
object_t reg_expr;
object_t reg_env;
object_t reg_cont;
object_t reg_val;
object_t reg_erator;
object_t reg_erand;

static void
invoke (void)
{
  switch ( obj_type (reg_cont) )
    {
    case O_CONT_FINAL:
      reg_task = new_task_final (reg_val);
      break;
    case O_CONT_EVAL:
      reg_task = new_task_eval (o_cont_eval_expr (reg_cont),
				o_cont_eval_env (reg_cont),
				o_cont_eval_cont (reg_cont));
      break;
    case O_CONT_APP1:
      reg_task = new_task_app1 (reg_val, o_cont_app1_rand (reg_cont),
				o_cont_app1_env (reg_cont),
				o_cont_app1_cont (reg_cont));
      break;
    case O_CONT_APP:
      reg_task = new_task_app (o_cont_app_erator (reg_cont), reg_val,
			       o_cont_app_cont (reg_cont));
      break;
    default:
      /* Something went wrong */
      error ("Uncaught case in invoke()");
      break;
    }
}

static object_t
env_lookup (hashv_t name, object_t env)
{
  while ( 1 )
    {
      switch ( obj_type (env) )
	{
	case O_ENV_EMPTY:
	  error ("Unbound variable");
	  break;
	case O_ENV_BINDING:
	  if ( o_env_binding_name (env) == name )
	    return o_env_binding_val (env);
	  env = o_env_binding_cdr (env);
	  break;
	default:
	  /* Something went wrong */
	  error ("Uncaught case in env_lookup()");
	  break;
	}
    }
}

static void
eval (void)
{
  switch ( obj_type (reg_expr) )
    {
    case O_QUOTE:
      reg_val = o_quote_val (reg_expr);
      invoke ();
      break;
    case O_VAR:
      reg_val = env_lookup (o_var_name (reg_expr), reg_env);
      invoke ();
      break;
    case O_PROGN:
      reg_cont = new_cont_eval (o_progn_pares (reg_expr),
				reg_env, reg_cont);
      reg_task = new_task_eval (o_progn_primus (reg_expr),
				reg_env, reg_cont);
      break;
    case O_APPLY:
      reg_cont = new_cont_app1 (o_apply_rand (reg_expr),
				reg_env, reg_cont);
      reg_task = new_task_eval (o_apply_rator (reg_expr),
				reg_env, reg_cont);
      break;
    case O_LAMBDA:
      reg_val = new_userfunc (o_lambda_param (reg_expr),
			      o_lambda_body (reg_expr), reg_env);
      invoke ();
      break;
    default:
      /* Something went wrong */
      error ("Uncaught case in eval()");
      break;
    }
}

static void
apply (void)
{
  switch ( obj_type (reg_erator) )
    {
    case O_USERFUNC:
      reg_env = new_env_binding (o_userfunc_param (reg_erator),
				 reg_erand, o_userfunc_defenv (reg_erator));
      reg_expr = o_userfunc_body (reg_erator);
      eval ();
      break;
    default:
      /* Something went wrong */
      error ("Uncaught case in apply()");
      break;
    }
}

void
run (void)
{
  switch ( obj_type (reg_task) )
    {
    case O_TASK_FINAL:
      error ("Program execution terminated");
      break;
    case O_TASK_EVAL:
      reg_expr = o_task_eval_expr (reg_task);
      reg_env = o_task_eval_env (reg_task);
      reg_cont = o_task_eval_cont (reg_task);
      eval ();
      break;
    case O_TASK_APP1:
      reg_expr = o_task_app1_rand (reg_task);
      reg_env = o_task_app1_env (reg_task);
      reg_cont = new_cont_app (o_task_app1_erator (reg_task),
			       o_task_app1_cont (reg_task));
      eval ();
      break;
    case O_TASK_APP:
      reg_erator = o_task_app_erator (reg_task);
      reg_erand = o_task_app_erand (reg_task);
      reg_cont = o_task_app_cont (reg_task);
      apply ();
      break;
    default:
      /* Something went wrong */
      error ("Uncaught case in run()");
      break;
    }
}
